"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ReliableRedisClient = void 0;
// Wraps redis client functionality by setting up a auto reconnecting redis client with warnings reporting to console
// Centralises reconnection strategy and gives better debugging and event handling
// Note: ioredis is leaps and bounds better than node-redis when it comes to handling unexpected disconnections
const ioredis_1 = __importDefault(require("ioredis"));
class ReliableRedisClient {
    constructor(connectionName, redisHost, redisPort) {
        this.connectionName = connectionName;
        this.redisHost = redisHost;
        this.redisPort = redisPort;
        this.client = null;
        this.isStarting = false;
        this.isStarted = false;
        this.isStopped = false;
        this.connectionAttempt = 0;
        this.lastConnectedAt = null;
        this.lastDisconnectAt = null;
        this.disconnectWarning = false; // This keeps track of whether a 15 second disconnect debug message has been sent
    }
    getNewRedisClient() {
        const client = new ioredis_1.default({
            host: this.redisHost,
            port: this.redisPort,
            autoResendUnfulfilledCommands: false,
            autoResubscribe: false,
            enableOfflineQueue: true,
            commandQueue: true,
            lazyConnect: true,
            // Note: Retrying forever sounds like the most recoverable option but is not very useful in reality.  
            // It can cause processing processes to remain alive but in a hung state forever.
            // It is better for commands to fail and processes to die so potential issues don't go unnoticed.
            // Note: This is the maximum connection retries per command
            maxRetriesPerRequest: 10,
            // Note: This is the retry connection strategy
            retryStrategy(times) {
                const delay = Math.min(times * 500, 5000);
                return delay;
            },
        });
        return client;
    }
    start(eventsCallback) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isStopped) {
                throw new Error("Is already stopped, cannot start");
            }
            if (this.isStarting) {
                throw new Error("Already started up by something else");
                // return false; // Already started up by something else
            }
            this.isStarting = true;
            this.client = this.getNewRedisClient();
            this.client.on('error', (e) => {
                if (e.code === "ECONNREFUSED") {
                    // Ignore
                }
                else {
                    console.warn("[Redis: " + this.connectionName + "] Error - " + e);
                }
            });
            this.client.on('connect', () => {
                // console.log("REDIS CLIENT " + this.connectionId + ": Connected");
            });
            this.client.on('ready', () => {
                const now = new Date();
                if (this.lastDisconnectAt !== null) {
                    const duration = Math.round((now.getTime() - this.lastDisconnectAt.getTime()) / 1000);
                    console.log("[Redis: " + this.connectionName + "] Re-connected to Redis after being down for " + duration + " secs");
                    if (eventsCallback) {
                        eventsCallback({
                            type: "RECONNECTED",
                            wasDisconnectedForSecs: duration,
                        });
                    }
                }
                else {
                    console.log("[Redis: " + this.connectionName + "] Connected to Redis!");
                    if (eventsCallback) {
                        eventsCallback({
                            type: "FIRST_READY",
                        });
                    }
                }
                this.connectionAttempt = 0;
                this.disconnectWarning = false;
                this.lastDisconnectAt = null;
                this.lastConnectedAt = new Date();
            });
            this.client.on('close', () => {
                // Not an error, just the connection closed
                // console.error("REDIS CLIENT " + this.connectionId + ": Close");
            });
            this.client.on('connecting', () => {
                if (this.lastConnectedAt === null) {
                    this.connectionAttempt++;
                    if (eventsCallback) {
                        eventsCallback({
                            type: "CONNECTING",
                            attempt: this.connectionAttempt,
                        });
                    }
                }
            });
            this.client.on('reconnecting', () => {
                if (this.lastConnectedAt === null) {
                    return; // Never connected yet, ignore this one since connect event still fires properly
                }
                this.connectionAttempt++;
                if (this.lastDisconnectAt === null) {
                    this.lastDisconnectAt = new Date();
                    console.warn("[Redis: " + this.connectionName + "] Warning, we are disconnected from redis!");
                }
                const now = new Date();
                const timeReconnecting = Math.round((now.getTime() - this.lastDisconnectAt.getTime()) / 1000);
                if (timeReconnecting > 15) {
                    if (!this.disconnectWarning) {
                        this.disconnectWarning = true;
                        console.warn("[Redis: " + this.connectionName + "] Warning, been disconnected from redis for " + timeReconnecting + " secs..");
                    }
                }
                if (eventsCallback) {
                    eventsCallback({
                        type: "RECONNECTING",
                        attempt: this.connectionAttempt,
                        reconnectingForSecs: timeReconnecting,
                    });
                }
            });
            const timeoutMs = 30 * 1000;
            try {
                yield this.client.connect();
            }
            catch (e) {
                // First connection attempt failed, but it will retry
                console.warn("[Redis: " + this.connectionName + "] Warning, the first connection attempt to Redis " + this.redisHost + ":" + this.redisPort + " failed, but will continue to retry for " + timeoutMs + " ms...");
            }
            let hasTimedOut = false;
            const connectionTimeout = setTimeout(() => {
                hasTimedOut = true;
            }, timeoutMs);
            while (!hasTimedOut && (this.lastDisconnectAt !== null || this.lastConnectedAt === null)) {
                // Not connected yet
                // console.warn("Still no connected yet");
                yield this.sleep(1000);
            }
            if (hasTimedOut) {
                this.client.disconnect(false);
                throw new Error("Connection timeout occurred after: " + timeoutMs);
            }
            clearTimeout(connectionTimeout);
            this.isStarted = true;
        });
    }
    sleep(ms) {
        return new Promise((resolve, reject) => {
            setTimeout(() => {
                resolve(null);
            }, ms);
        });
    }
    getClient() {
        if (this.client === null) {
            throw new Error("You must startup this controller");
        }
        else {
            if (this.isStarted) {
                return this.client;
            }
            else {
                throw new Error("Redis client still starting up, please await the startup function properly");
            }
        }
    }
    stop() {
        if (this.isStopped) {
            throw new Error("Is already stopped");
        }
        if (!this.isStarted) {
            throw new Error("Not started up properly yet, please await properly");
        }
        const client = this.getClient();
        client.removeAllListeners();
        client.disconnect(false);
        this.isStopped = true;
    }
}
exports.ReliableRedisClient = ReliableRedisClient;
